# Font utilities
from struct import pack, unpack

class Glyph(object):
    def __init__(self):
        self.width = 0
        self.pixels = '\x00'*16*16
        self.blank_prefix = 16
        self.blank_suffix = 16
        self.real_char_size = 0
        self.width_index = 0
        self.value = ''
        
    def __add__(self, og):
        sumg = Glyph()
        sumg.width = self.width + og.width
        wa = self.width
        wb = og.width
        sumg.blank_prefix = min(self.blank_prefix, og.blank_prefix)
        sumg.blank_suffix = min(self.blank_suffix, og.blank_suffix)

        if self.real_char_size == 0 and og.real_char_size == 0:
            sumg.real_char_size = 0
            sumg.pixels = '\x00'*16*16
        else:
            sumg.real_char_size = 16*4-sumg.blank_prefix-sumg.blank_suffix
            sumg.pixels = ''
            for j in range(16):
                sumg.pixels += self.pixels[16*j:16*j+wa]+\
                    og.pixels[16*j:16*j+wb]+'\x00'*(16-sumg.width)
        sumg.value = self.value + og.value
        return sumg

    def serialize(self, idx=None):
        if idx is not None:
            info = (idx << 16) + 4
        else:
            info = (self.width_index << 16) + 4
        if self.real_char_size == 0:
            info += (0xff << 8)
        else:
            info += ((self.blank_prefix / 4) << 12) + ((self.blank_suffix / 4) << 8)  +\
                self.real_char_size
        ser = pack('<I', info)
#        print self.value, hex(info)
        for j in range(self.real_char_size):
            v = 0
            for k in range(4):
                v += (ord(self.pixels[4*self.blank_prefix+4*j+k]) << (2*k))
            ser += chr(v)
        return ser

    def load(self, font_table, font_width_table, load_lazy=False):
        glyph_info = unpack('<I',font_table.read(4))[0]                    
        self.width_index = glyph_info >> 16

#        print self.value, self.width_index, hex(glyph_info)

        self.width = font_width_table[self.width_index]

        if load_lazy:
            return

        b, a = (glyph_info >> 8) % 256, glyph_info % 256
        self.blank_prefix = ((b >> 4) & 0xf)*4
        self.blank_suffix = (b & 0xf)*4
        self.real_char_size = (a - 4) & 0xff

        if self.blank_prefix == 0xf*4 and self.blank_suffix == 0xf*4:
            self.pixels = '\x00'*16*16
            return

        self.pixels = ''
        self.pixels += '\x00'*self.blank_prefix*4

        for k in range(self.real_char_size):
            c = ord(font_table.read(1))
            for j in range(4):
                v = (c >> (2*j)) % 4
                self.pixels += chr(v)

        self.pixels += '\x00'*self.blank_suffix*4

    def load_image(self, fim, x, font_width_table):
        w = 0
        for j in range(16):
            if fim.getpixel((x+1+j,0)) == 4:
                break
            w += 1
        
        self.width = w
        font_width_table[self.width_index] = self.width

        blank_prefix = 0
        for j in range(16):
            blineempty = True
            for k in range(w):
                blineempty = blineempty and (fim.getpixel((x+1+k,j)) == 0)
            if not blineempty: break
            blank_prefix += 4

        s = '\x00'*blank_prefix*4

        blank_suffix = 0
        for j in range(16):
            blineempty = True
            for k in range(w):
                blineempty = blineempty and (fim.getpixel((x+1+k,15-j)) == 0)
            if not blineempty: break
            blank_suffix += 4

        if blank_prefix == blank_suffix and blank_prefix == 16*4:
            s = '\x00'*16*16
            real_char_size = 0
        else:
            for j in range(blank_prefix/4, 16-blank_suffix/4):
                ls = ''
                for k in range(w):
                    ls += chr(fim.getpixel((x+1+k,j)))
                s += ls+'\x00'*(16-w)

            real_char_size = 16*4-blank_suffix-blank_prefix
            s += '\x00'*blank_suffix*4
        
        self.pixels = s
        self.blank_suffix = blank_suffix
        self.blank_prefix = blank_prefix
        self.real_char_size = real_char_size

class FontTable(object):
    def __init__(self, font_table, font_width_table, lookup=None, load_lazy=False):
        self.font_width_table = font_width_table
        self.glyphs = []
        font_table.seek(0)
        nglyphs = unpack('<H', font_table.read(2))[0]  
        glyphsoffset = unpack('<'+'I'*nglyphs, font_table.read(4*nglyphs))

        for i in range(nglyphs):
            g = Glyph()
            if lookup:
                g.value = lookup(pack('<H', 0x2000+i))                
#            print hex(i+0x2000)
            g.load(font_table, self.font_width_table, load_lazy=load_lazy)
            self.glyphs.append(g)

    def load_image(self, fim):
        for i in range(len(self.glyphs)):
            v = self.glyphs[i].width
            self.glyphs[i].load_image(fim, 17*i, self.font_width_table)
            w = self.glyphs[i].width

    def serialize(self, f, doublel=[], tokeep=[]):
        f.write(pack('<H', len(self.glyphs)))
        base_off = 2+4*len(self.glyphs)
        ft = ''

        bcomp = doublel != []
        doublep = 0
        doubleg = {}
        fwt = []

        for i in range(len(self.glyphs)):
            f.write(pack('<I', base_off+len(ft)))
#            print hex(i+0x2000)
            if bcomp and i not in tokeep and doublep < len(doublel):
                (a, b), c = doublel[doublep]
                doubleg[(a,b)] = i
                sum = self.glyphs[a-0x2000]+self.glyphs[b-0x2000]

                if sum.width in fwt:
                    k = fwt.index(sum.width)
                else:
                    k = len(fwt)
                    fwt.append(sum.width)

                ft += sum.serialize(idx=k)
                doublep += 1
            else:
                if self.glyphs[i].width in fwt:
                    k = fwt.index(self.glyphs[i].width)
                else:
                    k = len(fwt)
                    fwt.append(self.glyphs[i].width)
                ft += self.glyphs[i].serialize(idx=k)

        f.write(ft)

        return doubleg, fwt
